
# convert method implementation class

from .operators import *

class Convertor:

    def __init__(self):
        pass

    def coco2csv(self):
        # TODO: convert coco to csv
        pass

    def coco2pascalvoc(self):
        # TODO: convert coco to pascal VOC
        pass

    def csv2coco(self):
        # TODO: convert .csv into coco
        pass

    def csv2pascalvoc(self):
        # TODO: convert .csv into pascal VOC
        pass

    def pascalvoc2coco(self):
        # TODO: convert pascal VOC into coco
        pass

    def pascalvoc2csv(self):
        # TODO: convert pascal VOC into .csv
        pass




# Instance Object for COCO annotation format

from abc import ABC
import json
import os
import logging
import pandas as pd
import random

# setup logger
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


from .operator import IOperator


class COCO(IOperator, ABC):

    def __init__(self, dataset):
        super().__init__(dataset)
        self._dataset = dataset

    def get_dataset(self):
        return self._dataset

    def describe(self):
        # TODO: coco file description outputs (to - superClass )

        pass

    def sample(self, numOfSamples):
        """
        choose set of images randomly and get bounding boxes of them
        :return: dictionary list of [{"image_id" : "name", "classes" : [], "categories" : []}]
        """
        numOfrecords, _ = self._dataset.shape
        rnd_numbers = sorted(random.sample(range(0, numOfrecords), numOfSamples))
        sample_df = self._dataset.iloc[rnd_numbers, :]
        image_list = list(sample_df.loc[:, "image_id"].values)
        image_paths = list(sample_df.loc[:, "path"].values)
        sampled_anns = self.annotations.loc[self.annotations.loc[:, "image_id"].isin(image_list), :]
        final_list = []
        for image_id, image_path in zip(image_list, image_paths):
            ann_for_image = sampled_anns.loc[sampled_anns.loc[:, "image_id"] == image_id,:]
            spares_list = ann_for_image.values.tolist()
            ordered_dict = self.__listGen(spares_list)
            ordered_dict["image_id"] = image_id
            ordered_dict["path"] = image_path
            final_list.append(ordered_dict)
        return final_list

    def extract(self, path: str):
        """
        all the annotations in the file convert into general dataframe object.
        :param path: string, relative / absolute path
        :return: generalize pandas.DataFrame type object.
        """
        if os.path.exists(path):
            with open(path) as fp:
                ann_data = json.load(fp)
            self.__updateDataset(ann_data["images"])
            self.__extractAnnotation(ann_data["annotations"])
            self.__extractClasses(ann_data["categories"])
        else:
            logger.error(f"Error: entered path <{path}> is invalid.")
        return

    def archive(self):
        # TODO: save coco annotation file in the given location
        pass

    def translate(self):
        # TODO: translate common schema into json compatible format.
        pass

    def __normalized2KITTI(self, box):
        """

        :param box: [X, Y, width, highest]
        :return: [(xmin, ymin), (xmax, ymax)]
        """
        o_x, o_y, o_width, o_height = box
        xmin = int(o_x - o_width / 2)
        ymin = int(o_y - o_height / 2)
        xmax = int(o_x + o_width / 2)
        ymax = int(o_y + o_height / 2)
        return [(xmin, ymin), (xmax, ymax)]

    def __updateDataset(self, images):
        """

        :param images: image attributes in the .json file
        :return: add id, image width & height columns to self.dataset
        """
        dataset_imgs = list(self._dataset.iloc[:, 0].values)
        ann_imgs = []
        ann_id = {}
        img_width = {}
        img_height = {}
        for obj in images:
            if obj["file_name"] in dataset_imgs:
                try:
                    ann_imgs.append(obj["file_name"])
                    ann_id[obj["file_name"]] = obj["id"]
                    img_width[obj["file_name"]] = obj["width"]
                    img_height[obj["file_name"]] = obj["height"]
                except Exception as error:
                    logger.exception("ERROR: annotation file doesn't in accept the format.")
        if len(dataset_imgs) > len(ann_imgs):
            self._dataset = self._dataset.loc[self._dataset.loc[:, "name"].isin(ann_imgs), :]
            logger.warning("WARNING: all the images had not annotated!")
        self._dataset = self._dataset.copy()
        self._dataset["image_id"] = self._dataset["name"].map(ann_id)
        self._dataset.loc[:, "width"] = self._dataset.loc[:, "name"].map(img_width)
        self._dataset.loc[:, "height"] = self._dataset.loc[:, "name"].map(img_height)
        return

    def __extractAnnotation(self, anns):
        """

        :param anns: annotation attribute in the .json file
        :return: None , add self.annotations attr.
        """
        ann_list = []
        for obj in anns:
            try:
                obj_id = obj["id"]
                img_id = obj["image_id"]
                cls_id = obj["category_id"]
                min_tup, max_tup = self.__normalized2KITTI(obj["bbox"])
            except Exception as error:
                logger.exception("ERROR: annotation file doesn't in accept the format.")
            else:
                ann_list.append((obj_id, img_id, cls_id, min_tup[0], min_tup[1], max_tup[0], max_tup[1]))
        else:
            if ann_list:
                ann_df = pd.DataFrame.from_records(ann_list,
                                                   columns=['obj_id', 'image_id', 'class_id', 'x_min', 'y_min', 'x_max',
                                                            'y_max'])
                self.annotations = ann_df
            else:
                self.annotations = pd.DataFrame()

    def __extractClasses(self, cats):
        """

        :param cats: categories attribute in the .json file
        :return: dictionary object of type {id : class-name}
        """
        if len(cats) > 0:
            class_dict = {}
            for obj in cats:
                try:
                    class_dict[obj["id"]] = obj["name"]
                except Exception as error:
                    logger.exception("ERROR: annotation file doesn't in accept the format.")
            else:
                self.classes = class_dict
        else:
            logger.error("There are no distinctive class definition in the annotation.")
            self.classes = {}

    def __listGen(self, data_list):
        """

        :param data_list: [obj_id, class_id, class_id , x_min, y_min, x_max, y_max]
        :return: two list {"classes" : [classes, ..] , "bbox" : [[(x_min, y_min), (x_max, y_max)], ..]}
        """
        bounding_boxes = []
        classes = []
        for obj in data_list:
            classes.append(obj[2])
            bounding_boxes.append([(obj[3], obj[4]), (obj[5], obj[6])])
        final_dict = {"classes": classes, "bbox": bounding_boxes}
        return final_dict



# -*- coding: utf-8 -*-
import logging

from .operators.ImgData import ImgData
from .operators import coco, csv, pascalvoc

# setup logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

"""
### obj_lis : attributes ###
{
    ["classes" : [int, ..],
    "bbox" : [[(x_min, y_min), (x_max, y_max)], ..],
    "image_id" : int,
    "path: : str], ..
}
"""


class Sample:

    @classmethod
    def show_samples(cls, data_path: str,
                     ann_path: str,
                     num_of_samples: int = 5,
                     ann_type: str = 'coco'):
        """render set of random images from dataset.

        :param data_path: relative path current folder, or absolute path to the main folder of the image dataset
        :param ann_path: relative path current folder, or absolute path to the main folder of the annotated file
        :param ann_type: one of type from ['coco', 'voc', 'csv', 'yolo']
        :param num_of_samples: number of sample images in integer format.
        :return: render sequence of images.
        """

        imgdataset = ImgData.extract(data_path)
        if ann_type == 'coco':
            obj = coco.COCO(imgdataset.dataset)
        elif ann_type == 'voc':
            obj = pascalvoc.PascalVOC(imgdataset.dataset)
        elif ann_type == 'csv':
            obj = csv.CSV(imgdataset.dataset)
        elif ann_type == 'yolo':
            obj = csv.IOperator(imgdataset.dataset)
        else:
            assert Exception(f"ERROR: {ann_type} is not a valid annotation type.")

        obj.extract(ann_path)
        obj_list = obj.sample(num_of_samples)
        cat_dict = obj.classes
        for img_obj in obj_list:
            path = img_obj["path"]
            obj_data = img_obj["bbox"]
            cat_name = [cat_dict[j] for j in img_obj["classes"]]
            obj.render(path, obj_data, cat_name)
        return



# Instance Object for COCO annotation format

from abc import ABC

from .operator import IOperator


class CSV(IOperator, ABC):

    def __init__(self):
        pass

    def describe(self):
        # TODO: csv file description outputs (to - superClass )
        pass

    def sample(self):
        # TODO: output:: list of annotation results from annotation file
        pass

    def extract(self):
        # TODO: output:: all the annotations in the file
        pass

    def archive(self):
        # TODO: save csv annotation file in the given location
        pass

    def translate(self):
        # TODO: translate common schema into json compatible format.
        pass



# Operator Abstract class

from abc import ABCMeta, abstractmethod
import matplotlib.pyplot as plt
import cv2
import pandas as pd
import logging

# set logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

""":param
ann_df attributes:
    - obj_id : int
    - image_id : int
    - class_id : int
    - x_min : int
    - y_min : int
    - x_max : int
    - y_max : int
"""
""":param
render format
    - image path
    - box : [[(x_min, y_min), (x_max, y_max)], ...]
    - classes : [str, ...]
"""


class IOperator:

    def __init__(self, dataset):
        self.__dataset = dataset

    def set_dataset(self, df):
        """
        :param df: pandas.DataFrame type object with attr. defined in the ImgData.py file

        save new dataset into the objects' private variable.
        """
        if type(df) is pd.DataFrame:
            self.__dataset = df
        else:
            logger.error(f"Data type of df : {type(df)} not compatible with database object.")

    def get_dataset(self):
        """
        :return pandas.DataFrame
        """
        return self.__dataset

    @abstractmethod
    def describe(self):
        raise NotImplementedError

    @abstractmethod
    def sample(self, ann_data, names: list):
        raise NotImplementedError

    @abstractmethod
    def extract(self, path: str):
        raise NotImplementedError

    @abstractmethod
    def translator(self):
        raise NotImplementedError

    @abstractmethod
    def archive(self):
        raise NotImplementedError

    def descFormat(self):
        # TODO: make nice format to show descibe result.
        pass

    def render(self, path: str, boxes: list, cls: list, rect_th=1, text_size=0.5, text_th=1):
        # TODO: show annotated image
        img = cv2.imread(path)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        for i in range(len(boxes)):
            # print(boxes[i][0], boxes[i][1])
            cv2.rectangle(img, boxes[i][0], boxes[i][1], color=(14,14,14), thickness=rect_th)
            cv2.putText(img, cls[i],
                        boxes[i][0], cv2.FONT_HERSHEY_COMPLEX,
                        text_size, color=(1,1,1), thickness=text_th)
        plt.figure(figsize=(30, 30))
        plt.imshow(img)
        plt.xticks([])
        plt.yticks([])
        plt.show()
        return

    @classmethod
    def datasetReader(cls, data_path: str):

        return cls(object)

    @classmethod
    def randomizer(cls, num_of_samples: int):
        pass
